@@ -1,43 +1,42 @@
-# -*- coding=utf-8 -*-
-from __future__ import absolute_import, print_function
-
 import contextlib
 import importlib
-import io
+import itertools
 import json
 import operator
 import os
 import site
 import sys
-
+from pathlib import Path
 from sysconfig import get_paths, get_python_version
 
-import itertools
 import pkg_resources
-import six
 
 import pipenv
 
-from .vendor.cached_property import cached_property
-from .vendor.packaging.utils import canonicalize_name
-from .vendor import vistir
-
-from .utils import normalize_path, make_posix
+from pipenv.environments import is_type_checking
+from pipenv.utils import make_posix, normalize_path, subprocess_run
+from pipenv.vendor import vistir
+from pipenv.vendor.cached_property import cached_property
+from pipenv.vendor.packaging.utils import canonicalize_name
 
 
-if False:
+if is_type_checking():
+    from types import ModuleType
+    from typing import (
+        ContextManager, Dict, Generator, List, Optional, Set, Union
+    )
+
     import pip_shims.shims
     import tomlkit
-    from typing import ContextManager, Dict, Generator, List, Optional, Set, Union
-    from types import ModuleType
-    from .project import TSource, TPipfile, Project
-    from .vendor.packaging.version import Version
+
+    from pipenv.project import Project, TPipfile, TSource
+    from pipenv.vendor.packaging.version import Version
 
 BASE_WORKING_SET = pkg_resources.WorkingSet(sys.path)
 # TODO: Unittests for this class
 
 
-class Environment(object):
+class Environment:
     def __init__(
         self,
         prefix=None,  # type: Optional[str]
@@ -48,13 +47,13 @@ class Environment(object):
         sources=None,  # type: Optional[List[TSource]]
         project=None  # type: Optional[Project]
     ):
-        super(Environment, self).__init__()
+        super().__init__()
         self._modules = {'pkg_resources': pkg_resources, 'pipenv': pipenv}
         self.base_working_set = base_working_set if base_working_set else BASE_WORKING_SET
         prefix = normalize_path(prefix)
         self._python = None
         if python is not None:
-            self._python = vistir.compat.Path(python).absolute().as_posix()
+            self._python = Path(python).absolute().as_posix()
         self.is_venv = is_venv or prefix != normalize_path(sys.prefix)
         if not sources:
             sources = []
@@ -67,7 +66,7 @@ class Environment(object):
         self.pipfile = pipfile
         self.extra_dists = []
         prefix = prefix if prefix else sys.prefix
-        self.prefix = vistir.compat.Path(prefix)
+        self.prefix = Path(prefix)
         self._base_paths = {}
         if self.is_venv:
             self._base_paths = self.get_paths()
@@ -87,14 +86,6 @@ class Environment(object):
             if dist:
                 dist.activate()
             module = importlib.import_module(name)
-        if name in sys.modules:
-            try:
-                six.moves.reload_module(module)
-                six.moves.reload_module(sys.modules[name])
-            except TypeError:
-                del sys.modules[name]
-                sys.modules[name] = self._modules[name]
-                return self._modules[name]
         return module
 
     @classmethod
@@ -115,10 +106,15 @@ class Environment(object):
         try:
             reqs = dist.requires()
         # KeyError = limited metadata can be found
-        except (KeyError, AttributeError, OSError, IOError):  # The METADATA file can't be found
+        except (KeyError, AttributeError, OSError):  # The METADATA file can't be found
             return deps
         for req in reqs:
-            dist = working_set.find(req)
+            try:
+                dist = working_set.find(req)
+            except pkg_resources.VersionConflict:
+                # https://github.com/pypa/pipenv/issues/4549
+                # The requirement is already present with incompatible version.
+                continue
             deps |= cls.resolve_dist(dist, working_set)
         return deps
 
@@ -143,7 +139,7 @@ class Environment(object):
             return py_version
 
     def find_libdir(self):
-        # type: () -> Optional[vistir.compat.Path]
+        # type: () -> Optional[Path]
         libdir = self.prefix / "lib"
         return next(iter(list(libdir.iterdir())), None)
 
@@ -157,7 +153,7 @@ class Environment(object):
                 include_path = include_dirs.get("include", include_dirs.get("platinclude"))
                 if not include_path:
                     return {}
-                include_dir = vistir.compat.Path(include_path)
+                include_dir = Path(include_path)
         python_path = next(iter(list(include_dir.iterdir())), None)
         if python_path and python_path.name.startswith("python"):
             python_version = python_path.name.replace("python", "")
@@ -274,11 +270,11 @@ class Environment(object):
         if self._python is not None:
             return self._python
         if os.name == "nt" and not self.is_venv:
-            py = vistir.compat.Path(self.prefix).joinpath("python").absolute().as_posix()
+            py = Path(self.prefix).joinpath("python").absolute().as_posix()
         else:
-            py = vistir.compat.Path(self.script_basedir).joinpath("python").absolute().as_posix()
+            py = Path(self.script_basedir).joinpath("python").absolute().as_posix()
         if not py:
-            py = vistir.compat.Path(sys.executable).as_posix()
+            py = Path(sys.executable).as_posix()
         self._python = py
         return py
 
@@ -293,7 +289,7 @@ class Environment(object):
         """
 
         from .vendor.vistir.compat import JSONDecodeError
-        current_executable = vistir.compat.Path(sys.executable).as_posix()
+        current_executable = Path(sys.executable).as_posix()
         if not self.python or self.python == current_executable:
             return sys.path
         elif any([sys.prefix == self.prefix, not self.is_venv]):
@@ -331,20 +327,20 @@ class Environment(object):
         sysconfig_line = "sysconfig.get_path('{0}')"
         if python_lib:
             for key, var, val in (("pure", "lib", "0"), ("plat", "lib", "1")):
-                dist_prefix = "{0}lib".format(key)
+                dist_prefix = f"{key}lib"
                 # XXX: We need to get 'stdlib' or 'platstdlib'
-                sys_prefix = "{0}stdlib".format("" if key == "pure" else key)
-                pylib_lines.append("u'%s': u'{{0}}'.format(%s)" % (dist_prefix, distutils_line.format(var, val)))
-                pylib_lines.append("u'%s': u'{{0}}'.format(%s)" % (sys_prefix, sysconfig_line.format(sys_prefix)))
+                sys_prefix = "{}stdlib".format("" if key == "pure" else key)
+                pylib_lines.append(f"u'{dist_prefix}': u'{{{{0}}}}'.format({distutils_line.format(var, val)})")
+                pylib_lines.append(f"u'{sys_prefix}': u'{{{{0}}}}'.format({sysconfig_line.format(sys_prefix)})")
         if python_inc:
             for key, var, val in (("include", "inc", "0"), ("platinclude", "inc", "1")):
-                pylib_lines.append("u'%s': u'{{0}}'.format(%s)" % (key, distutils_line.format(var, val)))
+                pylib_lines.append(f"u'{key}': u'{{{{0}}}}'.format({distutils_line.format(var, val)})")
         lines = pylib_lines + pyinc_lines
         if scripts:
             lines.append("u'scripts': u'{{0}}'.format(%s)" % sysconfig_line.format("scripts"))
         if py_version:
             lines.append("u'py_version_short': u'{{0}}'.format(distutils.sysconfig.get_python_version()),")
-        lines_as_str = u",".join(lines)
+        lines_as_str = ",".join(lines)
         py_command = py_command % lines_as_str
         return py_command
 
@@ -361,12 +357,10 @@ class Environment(object):
         tmpfile_path = make_posix(tmpfile.name)
         py_command = self.build_command(python_lib=True, python_inc=True, scripts=True, py_version=True)
         command = [self.python, "-c", py_command.format(tmpfile_path)]
-        c = vistir.misc.run(
-            command, return_object=True, block=True, nospin=True, write_to_stdout=False
-        )
+        c = subprocess_run(command)
         if c.returncode == 0:
             paths = {}
-            with io.open(tmpfile_path, "r", encoding="utf-8") as fh:
+            with open(tmpfile_path, "r", encoding="utf-8") as fh:
                 paths = json.load(fh)
             if "purelib" in paths:
                 paths["libdir"] = paths["purelib"] = make_posix(paths["purelib"])
@@ -375,8 +369,8 @@ class Environment(object):
                     paths[key] = make_posix(paths[key])
             return paths
         else:
-            vistir.misc.echo("Failed to load paths: {0}".format(c.err), fg="yellow")
-            vistir.misc.echo("Output: {0}".format(c.out), fg="yellow")
+            vistir.misc.echo(f"Failed to load paths: {c.stderr}", fg="yellow")
+            vistir.misc.echo(f"Output: {c.stdout}", fg="yellow")
         return None
 
     def get_lib_paths(self):
@@ -391,13 +385,11 @@ class Environment(object):
         tmpfile_path = make_posix(tmpfile.name)
         py_command = self.build_command(python_lib=True)
         command = [self.python, "-c", py_command.format(tmpfile_path)]
-        c = vistir.misc.run(
-            command, return_object=True, block=True, nospin=True, write_to_stdout=False
-        )
+        c = subprocess_run(command)
         paths = None
         if c.returncode == 0:
             paths = {}
-            with io.open(tmpfile_path, "r", encoding="utf-8") as fh:
+            with open(tmpfile_path, "r", encoding="utf-8") as fh:
                 paths = json.load(fh)
             if "purelib" in paths:
                 paths["libdir"] = paths["purelib"] = make_posix(paths["purelib"])
@@ -406,8 +398,8 @@ class Environment(object):
                     paths[key] = make_posix(paths[key])
             return paths
         else:
-            vistir.misc.echo("Failed to load paths: {0}".format(c.err), fg="yellow")
-            vistir.misc.echo("Output: {0}".format(c.out), fg="yellow")
+            vistir.misc.echo(f"Failed to load paths: {c.stderr}", fg="yellow")
+            vistir.misc.echo(f"Output: {c.stdout}", fg="yellow")
         if not paths:
             if not self.prefix.joinpath("lib").exists():
                 return {}
@@ -445,20 +437,18 @@ class Environment(object):
             "fh = io.open('{0}', 'w'); fh.write(value); fh.close()"
         )
         command = [self.python, "-c", py_command.format(tmpfile_path)]
-        c = vistir.misc.run(
-            command, return_object=True, block=True, nospin=True, write_to_stdout=False
-        )
+        c = subprocess_run(command)
         if c.returncode == 0:
             paths = []
-            with io.open(tmpfile_path, "r", encoding="utf-8") as fh:
+            with open(tmpfile_path, "r", encoding="utf-8") as fh:
                 paths = json.load(fh)
             for key in ("include", "platinclude"):
                 if key in paths:
                     paths[key] = make_posix(paths[key])
             return paths
         else:
-            vistir.misc.echo("Failed to load paths: {0}".format(c.err), fg="yellow")
-            vistir.misc.echo("Output: {0}".format(c.out), fg="yellow")
+            vistir.misc.echo(f"Failed to load paths: {c.stderr}", fg="yellow")
+            vistir.misc.echo(f"Output: {c.stdout}", fg="yellow")
         return None
 
     @cached_property
@@ -472,8 +462,8 @@ class Environment(object):
         """
 
         command = [self.python, "-c", "import sys; print(sys.prefix)"]
-        c = vistir.misc.run(command, return_object=True, block=True, nospin=True, write_to_stdout=False)
-        sys_prefix = vistir.compat.Path(vistir.misc.to_text(c.out).strip()).as_posix()
+        c = subprocess_run(command)
+        sys_prefix = Path(c.stdout.strip()).as_posix()
         return sys_prefix
 
     @cached_property
@@ -525,7 +515,7 @@ class Environment(object):
         reinstall
         """
         prefixes = [
-            vistir.compat.Path(prefix)
+            Path(prefix)
             for prefix in self.base_paths["libdirs"].split(os.pathsep)
             if vistir.path.is_in_path(prefix, self.prefix.as_posix())
         ]
@@ -550,17 +540,15 @@ class Environment(object):
         :rtype: iterator
         """
 
-        pkg_resources = self.safe_import("pkg_resources")
         libdirs = self.base_paths["libdirs"].split(os.pathsep)
         dists = (pkg_resources.find_distributions(libdir) for libdir in libdirs)
-        for dist in itertools.chain.from_iterable(dists):
-            yield dist
+        yield from itertools.chain.from_iterable(dists)
 
     def find_egg(self, egg_dist):
         # type: (pkg_resources.Distribution) -> str
         """Find an egg by name in the given environment"""
         site_packages = self.libdir[1]
-        search_filename = "{0}.egg-link".format(egg_dist.project_name)
+        search_filename = f"{egg_dist.project_name}.egg-link"
         try:
             user_site = site.getusersitepackages()
         except AttributeError:
@@ -583,7 +571,7 @@ class Environment(object):
     def dist_is_in_project(self, dist):
         # type: (pkg_resources.Distribution) -> bool
         """Determine whether the supplied distribution is in the environment."""
-        from .project import _normalized
+        from .environments import normalize_pipfile_path as _normalized
         prefixes = [
             _normalized(prefix) for prefix in self.base_paths["libdirs"].split(os.pathsep)
             if _normalized(prefix).startswith(_normalized(self.prefix.as_posix()))
@@ -607,15 +595,12 @@ class Environment(object):
     @contextlib.contextmanager
     def get_finder(self, pre=False):
         # type: (bool) -> ContextManager[pip_shims.shims.PackageFinder]
-        from .vendor.pip_shims.shims import (
-            InstallCommand, get_package_finder
-        )
-        from .environments import PIPENV_CACHE_DIR
+        from .vendor.pip_shims.shims import InstallCommand, get_package_finder
 
         pip_command = InstallCommand()
         pip_args = self._modules["pipenv"].utils.prepare_pip_source_args(self.sources)
         pip_options, _ = pip_command.parser.parse_args(pip_args)
-        pip_options.cache_dir = PIPENV_CACHE_DIR
+        pip_options.cache_dir = self.project.s.PIPENV_CACHE_DIR
         pip_options.pre = self.pipfile.get("pre", pre)
         with pip_command._build_session(pip_options) as session:
             finder = get_package_finder(install_cmd=pip_command, options=pip_options, session=session)
@@ -623,7 +608,7 @@ class Environment(object):
 
     def get_package_info(self, pre=False):
         # type: (bool) -> Generator[pkg_resources.Distribution, None, None]
-        from .vendor.pip_shims.shims import pip_version, parse_version
+        from .vendor.pip_shims.shims import parse_version, pip_version
         dependency_links = []
         packages = self.get_installed_packages()
         # This code is borrowed from pip's current implementation
@@ -692,18 +677,19 @@ class Environment(object):
         return d
 
     def get_package_requirements(self, pkg=None):
-        from .vendor.pipdeptree import flatten, sorted_tree, build_dist_index, construct_tree
+        from .vendor.pipdeptree import PackageDAG, flatten
+
         packages = self.get_installed_packages()
         if pkg:
             packages = [p for p in packages if p.key == pkg]
-        dist_index = build_dist_index(packages)
-        tree = sorted_tree(construct_tree(dist_index))
-        branch_keys = set(r.key for r in flatten(tree.values()))
+
+        tree = PackageDAG.from_pkgs(packages).sort()
+        branch_keys = {r.key for r in flatten(tree.values())}
         if pkg is not None:
             nodes = [p for p in tree.keys() if p.key == pkg]
         else:
             nodes = [p for p in tree.keys() if p.key not in branch_keys]
-        key_tree = dict((k.key, v) for k, v in tree.items())
+        key_tree = {k.key: v for k, v in tree.items()}
 
         return [self._get_requirements_for_package(p, key_tree) for p in nodes]
 
@@ -722,7 +708,7 @@ class Environment(object):
         yield new_node
 
     def reverse_dependencies(self):
-        from vistir.misc import unnest, chunked
+        from vistir.misc import chunked, unnest
         rdeps = {}
         for req in self.get_package_requirements():
             for d in self.reverse_dependency(req):
@@ -745,9 +731,9 @@ class Environment(object):
         for k in list(rdeps.keys()):
             entry = rdeps[k]
             if entry.get("parents"):
-                rdeps[k]["parents"] = set([
+                rdeps[k]["parents"] = {
                    p for p, version in chunked(2, unnest(entry["parents"]))
-                ])
+                }
         return rdeps
 
     def get_working_set(self):
@@ -790,6 +776,8 @@ class Environment(object):
                     vcs_type == req.vcs and commit_id == req.commit_hash
                     and direct_url_metadata["url"] == pipfile_part[req.vcs]
                 )
+            elif req.is_vcs or req.is_file_or_url:
+                return False
             elif req.line_instance.specifiers is not None:
                 return req.line_instance.specifiers.contains(
                     match.version, prereleases=True
@@ -814,7 +802,7 @@ class Environment(object):
         return c
 
     def run_py(self, cmd, cwd=os.curdir):
-        """Run a python command in the enviornment context.
+        """Run a python command in the environment context.
 
         :param cmd: A command to run in the environment - runs with `python -c`
         :type cmd: str or list
@@ -824,8 +812,8 @@ class Environment(object):
         """
 
         c = None
-        if isinstance(cmd, six.string_types):
-            script = vistir.cmdparse.Script.parse("{0} -c {1}".format(self.python, cmd))
+        if isinstance(cmd, str):
+            script = vistir.cmdparse.Script.parse(f"{self.python} -c {cmd}")
         else:
             script = vistir.cmdparse.Script.parse([self.python, "-c"] + list(cmd))
         with self.activated():
@@ -837,8 +825,8 @@ class Environment(object):
         if self.is_venv:
             activate_this = os.path.join(self.scripts_dir, "activate_this.py")
             if not os.path.isfile(activate_this):
-                raise OSError("No such file: {0!s}".format(activate_this))
-            with open(activate_this, "r") as f:
+                raise OSError(f"No such file: {activate_this!s}")
+            with open(activate_this) as f:
                 code = compile(f.read(), activate_this, "exec")
                 exec(code, dict(__file__=activate_this))
 
@@ -867,7 +855,7 @@ class Environment(object):
             extra_dists = []
         original_path = sys.path
         original_prefix = sys.prefix
-        parent_path = vistir.compat.Path(__file__).absolute().parent
+        parent_path = Path(__file__).absolute().parent
         vendor_dir = parent_path.joinpath("vendor").as_posix()
         patched_dir = parent_path.joinpath("patched").as_posix()
         parent_path = parent_path.as_posix()
@@ -881,12 +869,11 @@ class Environment(object):
             ])
             os.environ["PYTHONIOENCODING"] = vistir.compat.fs_str("utf-8")
             os.environ["PYTHONDONTWRITEBYTECODE"] = vistir.compat.fs_str("1")
-            from .environments import PIPENV_USE_SYSTEM
             if self.is_venv:
                 os.environ["PYTHONPATH"] = self.base_paths["PYTHONPATH"]
                 os.environ["VIRTUAL_ENV"] = vistir.compat.fs_str(prefix)
             else:
-                if not PIPENV_USE_SYSTEM and not os.environ.get("VIRTUAL_ENV"):
+                if not self.project.s.PIPENV_USE_SYSTEM and not os.environ.get("VIRTUAL_ENV"):
                     os.environ["PYTHONPATH"] = self.base_paths["PYTHONPATH"]
                     os.environ.pop("PYTHONHOME", None)
             sys.path = self.sys_path
@@ -911,7 +898,6 @@ class Environment(object):
             finally:
                 sys.path = original_path
                 sys.prefix = original_prefix
-                six.moves.reload_module(pkg_resources)
 
     @cached_property
     def finders(self):
@@ -942,11 +928,11 @@ class Environment(object):
         install_args = [
             self.environment.python, "-u", "-c", SETUPTOOLS_SHIM % setup_path,
             install_arg, "--single-version-externally-managed", "--no-deps",
-            "--prefix={0}".format(self.base_paths["prefix"]), "--no-warn-script-location"
+            "--prefix={}".format(self.base_paths["prefix"]), "--no-warn-script-location"
         ]
         for key in install_keys:
             install_args.append(
-                "--install-{0}={1}".format(key, self.base_paths[key])
+                f"--install-{key}={self.base_paths[key]}"
             )
         return install_args
 
@@ -1022,7 +1008,7 @@ class Environment(object):
                 return
 
 
-class PatchedUninstaller(object):
+class PatchedUninstaller:
     def _permitted(self, path):
         return True
 
